#!/bin/sh --

cat <<EOF
/* this is an autogenerated file.  do not edit. */
#include <stdio.h>
#include <stdlib.h>

#include "mutt.h"

void mutt_print_patchlist (void)
{
EOF

while read patch ; do
	echo "  puts (\"${patch}\");"
done

echo "}"
